.editable-note-actions{margin-top:16px}.editable-note-actions button+button{margin-left:16px}.mde-header{background:var(--secondary-background-color)}.RequestStateWrapper-error{margin-left:auto;margin-right:auto;margin-top:50px;width:auto}