.ant-form-item{margin-bottom:8px}.ant-table-placeholder{padding:8px}.ant-modal-mask,.ant-modal-wrap div{animation-duration:0s!important}.fa{animation-duration:2s!important}.ant-dropdown-menu-item:hover,.ant-select-dropdown .ant-select-dropdown-menu-item-active,.ant-select-dropdown .ant-select-dropdown-menu-item:active,.ant-select-dropdown .ant-select-dropdown-menu-item:hover{background-color:#0193e1;background-image:none;color:#fff}.ant-dropdown-menu-item,.ant-select-dropdown .ant-select-dropdown-menu-item{transition:all 0s}.ant-table-placeholder{z-index:0}.ant-collapse-borderless{background-color:transparent}.ant-collapse-header{width:-moz-fit-content;width:fit-content}a{color:#2374bb}a:focus,a:hover{color:#005580}.ant-btn-primary{background-color:#2374bb;border-color:#2374bb}.ant-table-placeholder{color:#6b6b6b}.ant-radio-button-wrapper-checked:first-child,.ant-radio-button-wrapper-checked:not(.ant-radio-button-wrapper-disabled),.ant-tabs-nav .ant-tabs-tab-active{color:#2374bb}.ant-alert,.ant-form,.ant-form-item,.ant-table{font-size:13px!important}.editable-note-actions{margin-top:16px}.editable-note-actions button+button{margin-left:16px}.mde-header{background:var(--secondary-background-color)}.collapsible-section.ant-collapse>.ant-collapse-item>.ant-collapse-header{align-items:center;display:flex;font-size:16px;padding-left:0;padding-right:0}.collapsible-section .ant-collapse-content-box{padding-left:0;padding-right:0}.collapsible-section .ant-collapse-item{border:0;overflow:hidden}.far,.fas{-webkit-font-smoothing:antialiased;-moz-osx-font-smoothing:grayscale;display:inline-block;font:normal normal normal 14px/1 FontAwesome;font-size:inherit;text-rendering:auto}.model-version-table{margin-bottom:25px}.model-version-table .model-version-link{cursor:pointer}.model-version-table .model-version-status-icon{font-size:13px}.model-version-table .ant-table-thead>tr>th:not(:last-child):not(.ant-table-selection-column):not(.ant-table-row-expand-icon-cell):not([colspan]):before{display:none}.table-endpoint-container{max-width:200px;width:200px}.table-endpoint-text{display:block;overflow:hidden;text-overflow:ellipsis;white-space:nowrap}.status-alert{border-radius:2px;margin-bottom:16px}.status-alert .model-version-status-icon{margin-left:-3px}.status-alert.status-alert-info .icon-pending{color:green;margin-right:4px}.status-alert.status-alert-info{border-left:2px solid #3895d3}.status-alert.status-alert-error{border-left:2px solid red}.breadcrumb-header-wrapper{display:flex;justify-content:space-between;width:100%}.breadcrumb-header{align-items:center;color:#333;display:flex;font-size:18px;font-weight:700;margin-bottom:24px;margin-right:10px}.version-follow-icon{margin-left:auto}.ant-popover-content{max-width:500px}.stage-transition-dropdown .ant-tag{border-radius:4px;cursor:pointer}.model-version-update-form .ant-form-item{margin-bottom:0}.model-version-update-form .ant-form-item-label{text-align:left;width:100%}.model-version-update-form div[role=status][class^=css-]{margin-right:-.5px}.metadata-container{align-items:center;display:flex;flex-wrap:wrap}.metadata-entry{margin-bottom:16px}.icon-pending,.icon-ready{color:green}.icon-fail{color:red}body{margin:0;padding:0}#root{height:100%}.center{text-align:center}.error-image{display:block;margin:12% auto 60px}