.pagination-section{padding-bottom:30px}.pagination-dropdown .ant-dropdown-menu-item{text-align:center}.ant-alert-info .ant-alert-icon{color:#00b379}.table-tag-container{height:100%;max-width:100px;overflow:hidden;text-overflow:ellipsis;white-space:nowrap}.ModelListView-filter-dropdown{margin-bottom:24px;margin-top:0;min-width:800px}.search-input-tooltip .du-bois-light-popover-inner .du-bois-light-popover-inner-content{background-color:rgba(0,0,0,.75);border-radius:4px;color:#fff}.search-input-tooltip .du-bois-light-popover-arrow-content{background-color:rgba(0,0,0,.75)}