"use strict";(self.webpackChunk_mlflow_mlflow=self.webpackChunk_mlflow_mlflow||[]).push([[3835],{62448:function(r,e){class n{}n.mlflowServices={MODEL_REGISTRY:"Model Registry",EXPERIMENTS:"Experiments",MODEL_SERVING:"Model Serving",RUN_TRACKING:"Run Tracking",CLUSTER_MANAGEMENT:"Cluster Management",FILE_MANAGEMENT:"File Management"},e.A=n},26645:function(r,e,n){n.d(e,{X:function(){return y}});var t=n(31014);function o(r,e,n,t){Object.defineProperty(r,e,{get:n,set:t,enumerable:!0,configurable:!0})}o({},"ErrorBoundary",(()=>u));o({},"ErrorBoundaryContext",(()=>a));const a=(0,t.createContext)(null),s={didCatch:!1,error:null};class u extends t.Component{state=s;static getDerivedStateFromError(r){return{didCatch:!0,error:r}}resetErrorBoundary=(()=>{var r=this;return function(){const{error:e}=r.state;if(null!==e){for(var n=arguments.length,t=new Array(n),o=0;o<n;o++)t[o]=arguments[o];r.props.onReset?.({args:t,reason:"imperative-api"}),r.setState(s)}}})();componentDidCatch(r,e){this.props.onError?.(r,e)}componentDidUpdate(r,e){const{didCatch:n}=this.state,{resetKeys:t}=this.props;n&&null!==e.error&&function(){let r=arguments.length>0&&void 0!==arguments[0]?arguments[0]:[],e=arguments.length>1&&void 0!==arguments[1]?arguments[1]:[];return r.length!==e.length||r.some(((r,n)=>!Object.is(r,e[n])))}(r.resetKeys,t)&&(this.props.onReset?.({next:t,prev:r.resetKeys,reason:"keys"}),this.setState(s))}render(){const{children:r,fallbackRender:e,FallbackComponent:n,fallback:o}=this.props,{didCatch:s,error:u}=this.state;let i=r;if(s){const r={error:u,resetErrorBoundary:this.resetErrorBoundary};if((0,t.isValidElement)(o))i=o;else if("function"===typeof e)i=e(r);else{if(!n)throw new Error("react-error-boundary requires either a fallback, fallbackRender, or FallbackComponent prop");i=(0,t.createElement)(n,r)}}return(0,t.createElement)(a.Provider,{value:{didCatch:s,error:u,resetErrorBoundary:this.resetErrorBoundary}},i)}}function i(r){if(null==r||"boolean"!==typeof r.didCatch||"function"!==typeof r.resetErrorBoundary)throw new Error("ErrorBoundaryContext not found");return!0}function l(){const r=(0,t.useContext)(a);i(r);const[e,n]=(0,t.useState)({error:null,hasError:!1}),o=(0,t.useMemo)((()=>({resetBoundary:()=>{r?.resetErrorBoundary(),n({error:null,hasError:!1})},showBoundary:r=>n({error:r,hasError:!0})})),[r?.resetErrorBoundary]);if(e.hasError)throw e.error;return o}o({},"useErrorBoundary",(()=>l));function c(r,e){const n=n=>(0,t.createElement)(u,e,(0,t.createElement)(r,n)),o=r.displayName||r.name||"Unknown";return n.displayName=`withErrorBoundary(${o})`,n}o({},"withErrorBoundary",(()=>c));var d=n(82589),f=n(68254),h=n(88443),E=n(50111);function p(){return(0,E.Y)(d.SvL,{"data-testid":"fallback",title:(0,E.Y)(h.A,{id:"qAdWdK",defaultMessage:"Error"}),description:(0,E.Y)(h.A,{id:"RzZVxC",defaultMessage:"An error occurred while rendering this component."}),image:(0,E.Y)(f.k,{})})}function m(r){let{children:e}=r;return(0,E.Y)(u,{onError:function(r,e){console.error("Caught Unexpected Error: ",r,e.componentStack)},fallback:(0,E.Y)(p,{}),children:e})}function y(r,e,n){return function(r){return(0,E.Y)(m,{children:(0,E.Y)(e,{...r})})}}},25869:function(r,e,n){n.d(e,{h:function(){return a}});n(31014);var t=n(48624),o=n(50111);const a=r=>e=>{const n=(0,t.zy)(),a=(0,t.Zp)(),s=(0,t.g)();return(0,o.Y)(r,{params:s,location:n,navigate:a,...e})}},33835:function(r,e,n){n.r(e),n.d(e,{DirectRunPage:function(){return y},DirectRunPageImpl:function(){return p}});var t=n(82589),o=n(31014),a=n(10811),s=n(48624),u=n(76010),i=n(44240),l=n(58481),c=n(45395),d=n(25869),f=n(26645),h=n(62448),E=n(50111);const p=r=>{const{runUuid:e}=(0,s.g)(),[n,d]=(0,o.useState)(),f=(0,s.Zp)(),h=(0,a.wA)();return(0,o.useEffect)((()=>{d(void 0)}),[e]),(0,o.useEffect)((()=>{if(!r.runInfo&&e){const r=(0,i.aO)(e);r.payload.catch((r=>{u.A.logErrorAndNotifyUser(r),d(r)})),h(r)}}),[h,e,r.runInfo]),(0,o.useEffect)((()=>{var e;null!==(e=r.runInfo)&&void 0!==e&&e.experimentId&&f(l.h.getRunPageRoute(r.runInfo.experimentId,r.runInfo.runUuid),{replace:!0})}),[f,r.runInfo]),404===(null===n||void 0===n?void 0:n.status)?(0,E.Y)(c.o,{}):(0,E.Y)(t.ffj,{children:(0,E.Y)(t.PLz,{})})},m=(0,d.h)((0,a.Ng)(((r,e)=>({runInfo:r.entities.runInfosByUuid[e.params.runUuid]})))(p)),y=(0,f.X)(h.A.mlflowServices.RUN_TRACKING,m);e.default=y}}]);