.RequestStateWrapper-error{margin-left:auto;margin-right:auto;margin-top:50px;width:auto}.metrics-plot-container{align-items:flex-start;display:flex;width:100%}.metrics-plot-container .plot-controls{display:flex;flex-direction:column;min-height:500px}.metrics-plot-container .plot-controls .inline-control{align-items:center;display:flex;margin-top:25px}.metrics-plot-container .plot-controls .inline-control .control-label{margin-right:10px}.metrics-plot-container .plot-controls .block-control{margin-top:25px}.metrics-plot-container .metrics-plot-data{display:flex;flex:1 1;flex-direction:column}.metrics-plot-container .metrics-plot-view-container{flex:1 1;min-height:500px}.metrics-plot-container .metrics-summary{margin:20px 20px 20px 60px}.metrics-plot-container .metrics-summary .html-table-view{display:inline-block;margin-bottom:25px}.metrics-plot-container .metrics-summary .html-table-view th{min-width:auto}.html-table-view table{min-width:400px;width:auto}.html-table-view th{color:#888;font-size:13px;margin-right:80px;min-width:200px;width:auto}